<pre class='metadata'>
Title: CSS Extensions
Group: CSSWG
Shortname: css-extensions
Level: 1
Status: ED
Work Status: Exploring
ED: https://drafts.csswg.org/css-extensions
Editor: Tab Atkins, Google, http://xanthir.com/contact/, w3cid 42199
Abstract: This specification defines methods for authors to extend and enhance various CSS features.
Link Defaults: css-values-3 (dfn) identifier
</pre>

<h2 id='intro'>
Introduction</h2>

	When authoring CSS,
	one often encounters significant repetition in certain features.
	For example, a given media query might be repeated in several places,
	or a selector meant to apply to all heading elements
	requires specifying '':matches(h1, h2, h3, h4, h5, h6)'' in every location that uses it.

	This repetition makes stylesheets more verbose and difficult to read,
	and also affects maintenance,
	as the author has to keep each repetition in sync when making any changes.

	This specification defines methods for extending several CSS features
	so that a long or repeatedly-used value can be given a short, memorable name instead,
	or a feature can be given a more complex definition controlled by a scripting language.
	This makes stylesheets easier to read,
	and more powerful in general,
	as authors can extend the feature-set of CSS themselves
	rather than waiting for standards bodies to define new features for them.

<h2 id='extension-name'>
Extension Names</h2>

	All extensions defined in this specification use a common syntax for defining their ”names”:
	the <<extension-name>> production.
	An <dfn>&lt;extension-name></dfn> is any <a>identifier</a> that starts with two dashes (U+002D HYPHEN-MINUS),
	like ''--foo'', or even exotic names like ''--'' or ''------''.
	The CSS language will never use identifiers of this form for any language-defined purpose,
	so it's safe to use them for author-defined purposes
	without ever having to worry about colliding with CSS-defined names.

<h2 id='custom-selectors'>
<dfn lt="custom selector">Custom Selectors</dfn></h2>

	A <dfn>declarative custom selector</dfn> is defined with the ''@custom-selector'' rule:

	<pre class='prod'>
	<dfn>@custom-selector</dfn> = @custom-selector <<custom-selector>> <<selector-list>> ;
	<dfn>&lt;custom-selector></dfn> = <<custom-arg>>? : <<extension-name>> [ ( <<custom-arg>>+#? ) ]? ;
	<dfn>&lt;custom-arg></dfn> = $ <<ident-token>> ;
	</pre>

	Where there must be no whitespace
	between <code>:</code> and <<extension-name>>
	or between <code>$</code> and <<ident-token>>
	in the above definitions.

	<div class='issue'>
		Simple things are easy:

		<pre>
		@custom-selector :--heading {
			expansion: h1, h2, h3, h4, h5, h6;
		}
		</pre>

		More complicated things are possible:

		<pre>
		// Arguments are specified with $foo.
		// An arg before the pseudo-class captures the rest of the compound selector.
		@custom-selector $rest:--n-siblings($n, $sel) {
			specificity: $sel;
			// assumes $sel is a selector, parses it and uses its specificity
			// otherwise, specificity is [0,1,0]
			expansion: $rest:nth-child(1 of $sel):nth-last-child($n of $sel),
				:nth-child(1 of $sel):nth-last-child($n of $sel) ~ $rest;
		}
		</pre>
	</div>

	This defines a <a>custom selector</a> which is written as a <a spec=selectors>pseudo-class</a> with the given <<extension-name>>,
	and represents a '':matches()'' selector using the provided <<selector-list>> as its argument.

	<div class='example'>
		For example, if an author wanted to easily refer to all heading elements in their HTML document,
		they could create an alias:

		<pre>
			@custom-selector :--heading h1, h2, h3, h4, h5, h6;

			:--heading { /* styles for all headings */ }
			:--heading + p { /* more styles */ }
			/* etc */
		</pre>
	</div>

<h3 id='script-custom-selectors'>
Script-based Custom Selectors</h3>

	<div class='issue'>
		This one's more complicated than MQs.
		Brian Kardell came up with a good proposal for evaluating selectors as JS functions that return a boolean,
		which had decent performance characteristics by specifying the qualities of the element it was based on
		(which determined when it would be called).

		<pre>
		&lt;script>
		CSS.customSelector.set("_foo",
			                   {"predicate": function(el){...},
			                   	"matches": "a"});
		&lt;/script>
		</pre>

		"matches" is an optional selector specifying what subset of elements the custom selector is valid for.
		The selector is automatically false for elements that don't match,
		and the predicate isn't called.

		By default, the predicate is called whenever there's a mutation in an element that matches the "matches" selector,
		or one of its descendants.

		You should be able to suppress the auto-calling,
		and be able to trigger the predicate to run manually.
		That way you can use mutation listeners manually to only call the predicate when necessary.

		We should probably offer some sugar for filtering the list of mutations that trigger the predicate to be called.
		Maybe just a list of attributes that you'll be caring about?  And/or tagnames?

		Maybe let the pseudo-class also accept an argument,
		and pass it (as a serialized string) as a second argument to the predicate.
		'':_foo'' would pass <code>null</code>,
		while '':_foo()'' would pass <code>""</code>.
	</div>

<h3 id='custom-selectors-cssom'>
CSSOM</h3>

	<p class='issue'>
		Fill in.

<h2 id='custom-property'>
Custom Properties</h2>

	<div class='issue'>
		Need to more fully support Custom Properties
		(and eventually remove them from the variable spec entirely, since they'll be defined here).

		By default, custom properties are optimized for use as ''var()'' values--
		they inherit,
		have an empty initial value,
		don't do any syntax checking,
		and don't animate.
		All of these should be adjustable somehow.

		<pre>
		@custom-property --foo {
			scope: [ inherit | local ];
			initial: <<declaration-value>>*;
			value: <<length>> <<length>> <<color>>;
			/* Literally, define a simplistic definition syntax.
			   OR FULL CSS PROPERTY GRAMMAR?!? */
		}
		</pre>

		If you provide a "value" field with animatable types,
		we can animate in the most direct fashion automatically.
		We could also let you hook into that:
		you register a callback,
		and whenever a property starts animating,
		we call it with the starting and ending values.
		You have to return a function which takes a progress value (between 0 and 1)
		and returns a value for your property;
		we'll call it as we animate the value.
		(How can we hook into Web Anim here? Can you just return an Animation object?)

		Do we need a hook for computed values?  Interesting.
		We could just hand your callback a set of property values for the element and its parent (maybe siblings, if you ask for it?),
		and you can return a new value for the property.
		This is probably an advanced feature for a later date.

		Definitely need a way to listen for elements receiving and changing property values,
		so you can efficiently polyfill things and make your own properties.
		Unsure how it would look at the moment.
	</div>

<h2 id='custom-functions'>
Custom Functions</h2>

	<div class='issue'>
		Interesting possibilities here.
		Definitely need some way to define custom functions in CSS.
		This would, for example, let people define whatever color function they want,
		such as implementing the <a href="http://www.boronine.com/husl/">HUSL</a> color space.

		Definitely need a JS interface.
		What options are needed?

		Call time/frequency:

		<ul>
			<li>
				Default should probably treat the function as a preprocessor,
				calling the JS function once per instance in the stylesheet
				and substituting in the returned value.

			<li>
				Should probably have an option to allow calling per element/instance combo, too.
				Gets called more as match results change.
		</ul>

		We can take some cues from my thoughts on a random() function.
		It needs per-instance,
		per-element&amp;instance,
		and per "identifier", so you can reuse the same value in multiple spots.
		That last one can probably be handled manually by the JS,
		so we don't have to privilege a particular argument as an identifier.

		We'd need to provide the context in which it's used.
		Which property, for example.
		Should we allow them to be used in other places,
		or should we just define more contextual locations as we go?
		That is, should we allow custom-defined functions in @supports with this API,
		or should we add a <code>.customSupports</code> map?
		I suspect that individual cases will have their own useful contextual information,
		so it's better to specialize each instance of custom functions.

		How much can we do in pure CSS?
		Being able to substitute values depending on MQs or support queries would be useful.
		(However, we can do that much just by using custom properties and ''var()''.)
		To get *real* use out of it, though, I suspect we'd need fuller support for conditionals,
		likely in the form of SASS's ''@if'' or something similar.
	</div>

<h2 id='custom-combinators'>
Custom Selector Combinators</h2>

	<div class='issue'>
		Selectors are made of two pieces:
		simple selectors,
		and combinators.
		We should allow custom combinators too.

		This is JS-only, because it's transforming elements, not filtering them,
		and you can't express any useful transformations in pure CSS.

		You provide a function which,
		when given an element,
		produces a list of zero or more elements.

		For examples, with ''div /--foo/ span'',
		the CSS engine will match the first part of the selector
		and find all the div elements.
		It passes that list to the function registered for the --foo combinator,
		and expects to get a new list of elements returned.
		It then continues on its way,
		filtering that list to include only span elements, etc.

		A child combinator would be something like:

		<pre>
		CSS.customCombinator.set("--child", function(el) {
				return el.children;
			});
		</pre>

		Then ''div /--child/ span'' would be identical to ''div > span''.

		If we generalize a selector with a custom combinator to ''A /--custom/ B'',
		then the UA would automatically call the --custom function
		whenever new elements match ''A''.
		If elements stop matching ''A'',
		it won't bother;
		it'll just drop them from the result.

		Alternately, the function could take a list of elements
		(all the elements matching ''A'')
		and return a new list of elements.
		This would be a bit more complicated for the author,
		but would allow more variety in the types of combinators that could be defined,
		as you could define things that depend on the entire set of matched elements.
		For example, you could define ''A /nth 1/ B''
		to give only the first element from the set of ''A'' matches.

		(Maybe we allow both variants,
		since the per-element one is easier to optimize and program against,
		but the per-set one allows some useful stuff.)

		Similarly to custom pseudo-classes,
		we'd allow arguments,
		with them parsed eagerly per-instance
		and passed to the combinator function.

		If we do the per-element combinator function,
		we could potentially cache the results,
		so that it never needs to be called again for the same element.
		Possibly have a flag that turns off this behavior,
		so that you're guaranteed to be called again.
	</div>

<h2 id='custom-atrules'>
Custom At-Rules</h2>

	<div class='issue'>
		This one's even less developed,
		but it would be interesting to allow custom at-rules as well.
		It's definitely pure-JS as well.

		Unsure exactly what's best here.
		Possibly register a callback per rule,
		which is called with the prelude/contents of the at-rule?

		Should we do the callback approach,
		or just maintain a list of custom at-rules
		and let scripts parse them themselves?
		Unfortunately, the latter means we'd have to have a special mechanism to alert scripts
		when new at-rules get added or removed.

		For a lot of these at-rules, we may want a way to know when they're "applied"--
		when, according to the built-in at-rules like @media and @supports,
		the rule would be applied.
	</div>
